/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <psp_req.h>

int
psp_req_init (psp_req_t ** req, transaction_t * inc_tr)
{
  *req = (psp_req_t *) smalloc (sizeof (psp_req_t));
  if (*req == NULL)
    return -1;

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO1, NULL,
	       "Allocating psp_req ressource!\n"));

  (*req)->inc_tr = inc_tr;
  if (inc_tr != NULL)
    (*req)->request = inc_tr->orig_request;
  else
    (*req)->request = NULL;
  (*req)->response = NULL;
  (*req)->start = time (NULL);
  (*req)->timeout = 32;		/* 32 s is the default for a search location? */
  (*req)->owners = 0;
  (*req)->flag = 0;
  (*req)->uas_status = 0;
  (*req)->locations = NULL;
  /*  (*req)->plug_ctxs        = NULL; */
  (*req)->rsp_ctx = NULL;
  (*req)->next = NULL;
  (*req)->parent = NULL;
  return 0;
}

void
psp_req_free (psp_req_t * req)
{
  /*  plug_ctx_t *c; */
  location_t *loc;

  if (req == NULL)
    return;

  if (req->owners != 0)
    return;			/* delete later */

  if (req->inc_tr == NULL)
    {
    }
  else if (req->inc_tr->state == IST_TERMINATED
	   || req->inc_tr->state == NIST_TERMINATED)
    {
      transaction_free2 (req->inc_tr);
      sfree (req->inc_tr);
    }
  else
    {
      transaction_free (req->inc_tr);
      sfree (req->inc_tr);
    }

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO1, NULL,
	       "free psp_req ressouce!\n"));
  for (loc = req->locations; loc != NULL; loc = req->locations)
    {
      REMOVE_ELEMENT (req->locations, loc);
      location_free (loc);
      sfree (loc);
    }
  /*
     for (c=req->plug_ctxs; c!=NULL; c=req->plug_ctxs)
     {
     REMOVE_ELEMENT(req->plug_ctxs, c);
     plug_ctx_free(c);
     sfree(c);
     }
   */
  rsp_ctx_free (req->rsp_ctx);
  sfree (req->rsp_ctx);
  /* This is not a copy, just a pointer on tranasaction->orig_request
     msg_free (req->request);
     sfree (req->request);
   */
  /*
     msg_free (req->response);
     sfree (req->response);
   */
}

PPL_DECLARE (transaction_t *)
psp_req_get_incoming_transaction (psp_req_t * req)
{
  if (req == NULL)
    return NULL;
  return req->inc_tr;
}

PPL_DECLARE (int) psp_req_set_mode (psp_req_t * req, int mode)
{
  if (req == NULL)
    return -1;
  req->flag = (req->flag & 0xFF0) | mode;	/* so we delete the first 4 bits */
  return 0;
}

PPL_DECLARE (int) psp_req_set_state (psp_req_t * req, int state)
{
  if (req == NULL)
    return -1;
  req->flag = (req->flag & 0xF0F) | state;	/* so we keep only the first 4 bits */
  return 0;
}

PPL_DECLARE (int) psp_req_set_property (psp_req_t * req, int property)
{
  if (req == NULL)
    return -1;
  req->flag = (req->flag & 0x0FF) | property;	/* so we delete the first 4 bits */
  return 0;
}

PPL_DECLARE (int) psp_req_set_uas_status (psp_req_t * req, int status)
{
  if (req == NULL)
    return -1;
  req->uas_status = status;
  return 0;
}

PPL_DECLARE (int) psp_req_get_uas_status (psp_req_t * req)
{
  if (req == NULL)
    return -1;
  return req->uas_status;
}

int
psp_req_set_response_context (psp_req_t * req, rsp_ctx_t * ctx)
{
  if (req == NULL)
    return -1;
  req->rsp_ctx = ctx;
  return 0;
}

int
psp_req_take_ownership (psp_req_t * req)
{
  if (req == NULL)
    return -1;
  req->owners++;
  return 0;
}

int
psp_req_release_ownership (psp_req_t * req)
{
  if (req == NULL)
    return -1;
  req->owners--;
  return 0;
}

PPL_DECLARE (int) psp_req_is_expired (psp_req_t * req)
{
  time_t date;

  if (req == NULL)
    return -1;
  date = time (NULL);
  if (date - req->start > req->timeout)	/* context is expired! */
    return 0;
  return -1;
}

PPL_DECLARE (int) psp_req_add_timeout (psp_req_t * req, int add_sec)
{
  if (req == NULL)
    return -1;
  req->timeout = req->timeout + add_sec;
  return 0;
}

PPL_DECLARE (int) psp_req_get_timeout (psp_req_t * req)
{
  if (req == NULL)
    return -1;
  return req->timeout;
}

PPL_DECLARE (int) psp_req_set_timeout (psp_req_t * req, int timeout)
{
  if (req == NULL)
    return -1;
  req->timeout = timeout;
  return 0;
}

PPL_DECLARE (int) location_init (location_t ** loc, url_t * url, int expire)
{
  *loc = (location_t *) smalloc (sizeof (location_t));
  if (*loc == NULL)
    return -1;
  (*loc)->url = url;
  (*loc)->path = NULL;
  (*loc)->expire = expire;
  (*loc)->next = NULL;
  (*loc)->parent = NULL;
  return 0;
}

PPL_DECLARE (void) location_free (location_t * location)
{
  url_free (location->url);
  sfree (location->url);
  sfree (location->path);
}

PPL_DECLARE (int) location_set_url (location_t * location, url_t * url)
{
  if (location == NULL)
    return -1;
  location->url = url;
  return 0;
}

PPL_DECLARE (int) location_set_expires (location_t * location, int expire)
{
  if (location == NULL)
    return -1;
  location->expire = expire;
  return 0;
}

PPL_DECLARE (int) location_set_path (location_t * location, char *path)
{
  if (location == NULL)
    return -1;
  location->path = path;
  return 0;
}

PPL_DECLARE (int) location_clone (location_t * location, location_t ** loc)
{
  int i;
  url_t *url;

  *loc = NULL;
  if (location == NULL)
    return -1;
  i = url_clone (location->url, &url);
  if (i != 0)
    return -1;
  i = location_init (loc, url, location->expire);
  if (i != 0)
    goto lc_error1;
  return 0;
lc_error1:
  url_free (url);
  sfree (url);
  *loc = NULL;
  return -1;
}

/* useless in new architecture
sfp_branch_t *
psp_core_sfp_find_branch (psp_req_t * req, int osip_id)
{
  sfp_branch_t *sfp;

  if (req == NULL || req->rsp_ctx == NULL)
    return NULL;
  for (sfp = req->rsp_ctx->branches; sfp != NULL; sfp = sfp->next)
    {
      if (sfp->osip_id == osip_id)
        {
          return sfp;
        }
    }
  return NULL;
}
*/
